#!/usr/bin/perl -w
#use POSIX qw(ceil);

use strict 'refs';
use strict 'subs';
use strict;

#assumptions:
# One kind of CPU assumed
# Compute requirements of storage server daemons ignored
# Not yet accounting for network traffic
# Simple accounting for communication between tasks

my $EPSILON = 1e-10;
my $INFINITY = 1000000000000;
my $DONT_STORE_SERVERS = 1;
my $DEBUG = 0;
my $NUM_SAMPLES = 1;
my $SELECT_INIT_CONFIG = 1;

if ($#ARGV != 2 and $#ARGV != 3) {
	print STDERR "Usage: scc.pl [-s] <resources_file> <app_char_file> <slas_file>\n";
	exit;
}

my $silent_mode = 0;
my $k = 0;
if ($#ARGV >= 3) {
	if ($ARGV[0] ne "-s") {
		print STDERR "Unrecognized option $ARGV[0]\n";
		exit;
	}
	$silent_mode = 1;
	++$k;
}
my $res_file = $ARGV[$k++];
my $app_char_file = $ARGV[$k++];
my $sla_file = $ARGV[$k++];

sub debug {
	my ($str) = @_;

	if ($DEBUG) {
		print "$str";
	}
}

sub min {
	my ($x, $y) = @_;

	return ($x < $y ? $x : $y);
}

sub max {
	my ($x, $y) = @_;

	return ($x > $y ? $x : $y);
}

sub max3 {
	my ($x, $y, $z) = @_;

	return max($x, max($y, $z));
}

sub ceil {
	my ($x) = @_;

	return (($x - int($x) < $EPSILON) ? int($x) : (int($x) + 1));
}

my @storage_res = ();
my @compute_res = ();
my @servers_res = ();
my %bus_hash = ();
open FILE, $res_file;
while (<FILE>) {
	chomp;
	if (m/^\#/) {
		next;
	}

	my @f = split " ";
	if ($f[0] =~ m/^s:/) { #storage units
		my ($name_token, $persistence, $cap, $rate_r, $rate_w, $gap_r, $gap_w, $bus, $price, $power) = @f;
		my ($dummy, $name) = split ":", $name_token;

		my %hash = ();
		$hash{"name"} = $name;
		$hash{"persistence"} = $persistence;
		$hash{"cap"} = $cap;
		$hash{"rate_r"} = $rate_r;
		$hash{"rate_w"} = $rate_w;
		$hash{"gap_r"} = $gap_r;
		$hash{"gap_w"} = $gap_w;
		$hash{"bus"} = $bus;
		$hash{"price"} = $price;
		$hash{"power"} = $power;
		push @storage_res, \%hash;
	}
	elsif ($f[0] =~ m/^c:/) { #compute units
		if ($#compute_res != -1) {
			print STDERR "Model accepts only one kind of processor\n";
			exit;
		}

		my ($name_token, $clock, $price, $power) = @f;
		my ($dummy, $name) = split ":", $name_token;

		my %hash = ();
		$hash{"name"} = $name;
		$hash{"clock"} = $clock;
		$hash{"price"} = $price;
		$hash{"power"} = $power;
		push @compute_res, \%hash;
	}
	elsif ($f[0] =~ m/^m:/) { #mother boards
		my $name_token = $f[0];
		my $nw_limit = $f[1];
		my $price = $f[2];
		my $power = $f[3];
		my ($dummy, $name) = split ":", $name_token;

		my %hash = ();
		$hash{"name"} = $name;
		$hash{"nwlimit"} = $nw_limit;
		$hash{"price"} = $price;
		$hash{"power"} = $power;
		for (my $i = 4; $i <= $#f; ++$i) {
			my ($token, $res_name, $units_limit, $io_or_cache_limit) = split ":", $f[$i];
			$hash{$token}{$res_name} = [$units_limit, $io_or_cache_limit];
			if ($token eq "b") {
				$bus_hash{$res_name} = 1;
			}
			elsif ($token eq "c" and $res_name ne ${$compute_res[0]}{"name"}) {
				print STDERR "Unsupported CPU type $res_name\n";
				exit;
			}
		}
		push @servers_res, \%hash;
	}
	else {
		print STDERR "Unrecognized resource line: $_\n";
		exit;
	}
}
close FILE;
my @bus_res = keys %bus_hash;

#properties of datasets
my %cap_hash = ();
my %persistence_hash = ();
my %per_process_copy_hash = ();
my %multiplex_factor = ();
my %slowdown_factor = ();

#properties of tasks
my %memory_hash = ();
my %remote_dataset_hash = ();
my %process_locality_hash = ();
my %phase_hash = ();
my %process_hash = ();
my %phase_sets = ();

#properties of IO operations and interactions
my %dataset_io = ();
my %task_io = ();
my %task_compute = ();
my %task_ops = ();
my %parent_tasks = ();
my %child_tasks = ();
my $edge_id = 0;
my %io_times = ();

#properties of processes
my %process_sets = ();
my %max_cpu_util = ();

open FILE, $app_char_file;
while (<FILE>) {
	chomp;
	if (m/^\#/) {
		next;
	}

	my @f = split " ";

	if ($f[0] =~ m/^d:/) {
		my ($dummy, $dataset, $persistence, $size, $multiplex, $slowdown, $per_process_copy, $locality, $process) = split ":", $f[0];
		if ($#f != 0) {
			print STDERR "Dataset characterization with more than 1 token: $_\n";
			exit;
		}

		$cap_hash{$dataset} = $size;
		$persistence_hash{$dataset} = $persistence;
		$multiplex_factor{$dataset} = $multiplex;
		$slowdown_factor{$dataset} = $slowdown;
		$per_process_copy_hash{$dataset} = $per_process_copy;
		if ($locality == 1) {
			if (not(exists($process_locality_hash{$process}))) {
				$process_locality_hash{$process} = [];
			}
			push @{$process_locality_hash{$process}}, $dataset;
		}
		elsif ($locality == 0) {
			$remote_dataset_hash{$dataset} = 1;
		}
		elsif ($locality == 2) {
			$remote_dataset_hash{$dataset} = 2;
		}
		else {
			print STDERR "Unrecognized value $locality for locality flag\n";
			exit;
		}
	}
	elsif ($f[0] =~ m/^t:/) {
		my ($dummy, $task, $memory, $phase, $process) = split ":", $f[0];
		$memory_hash{$task} = $memory;
		$phase_hash{$task} = $phase;
		$process_hash{$task} = $process;
		if (not(exists($process_sets{$process}))) {
			$process_sets{$process} = [];
		}
		push @{$process_sets{$process}}, $task;
		if (not(exists($phase_sets{$phase}))) {
			$phase_sets{$phase} = [];
		}
		push @{$phase_sets{$phase}}, $task;
		$task_ops{$task} = [];

		for (my $i = 1; $i <= $#f; ++$i) {
			my @g = split ":", $f[$i];
			++$edge_id;
			if ($g[0] eq "C") {
				my ($token, $compute_time, $blocking) = @g;
				$task_compute{$task}{$edge_id} = [$compute_time, $blocking];
				push @{$task_ops{$task}}, ["compute", $edge_id, $blocking];
			}
			elsif ($g[0] eq "R" or $g[0] eq "W") {
				my ($io_type, $dataset, $record_size, $num_records, $parallel, $blocking, $seq) = @g;
				$task_io{$task}{$edge_id} = [$dataset, $io_type, $num_records, $record_size, $parallel, $blocking, $seq];
				push @{$task_ops{$task}}, ["io", $edge_id, $blocking];
				$dataset_io{$dataset}{$edge_id} = [$task, $io_type, $num_records, $record_size, $parallel, $blocking, $seq];
				debug "IO $edge_id: $task, $dataset, $num_records, $record_size, $blocking, $seq\n";
			}
			elsif ($g[0] eq "T") {
				my ($token, $dep_task, $num_calls, $parallel, $blocking) = @g;
				$child_tasks{$task}{$edge_id} = [$dep_task, $num_calls, $parallel, $blocking];
				push @{$task_ops{$task}}, ["dep", $edge_id, $blocking];
				if (not(exists($parent_tasks{$dep_task}))) {
					$parent_tasks{$dep_task} = [];
				}
				push @{$parent_tasks{$dep_task}}, [$task, $num_calls];
			}
			else {
				print STDERR "Unrecognized app characterization line: $_\n";
				exit;
			}
		}
	}
	else {
		print STDERR "Unrecognized app characterization line: $_\n";
		exit;
	}
}
close FILE;

my %sla_values = ();

open FILE, $sla_file;
while (<FILE>) {
	chomp;
	if (m/^\#/) {
		next;
	}

	my @f = split " ";
	my ($sla_type, $sla_value) = split ":", $f[0];
	my $sla_task = $f[1];

	if ($sla_type eq "L") {
		print STDERR "Latency SLAs not yet supported\n";
		exit;
	}
	elsif ($sla_type eq "T") {
		$sla_values{$sla_task} = $sla_value;
	}
	else {
		print STDERR "SLA needs to be either latency or throughput constraint\n";
		exit;
	}
}
close FILE;

sub compute_runtime {
	my ($task) = @_;

	my $r = 0;
	my $curr_stage_time = 0;
	foreach my $op (@{$task_ops{$task}}) {
		my $curr_op_time;
		my ($op_type, $op_id, $blocking) = @{$op};
		if ($op_type eq "io") {
			$curr_op_time = $io_times{$op_id};
		}
		elsif ($op_type eq "compute") {
			my ($compute_time, $b) = @{$task_compute{$task}{$op_id}};
			$curr_op_time = $compute_time;
		}
		elsif ($op_type eq "dep") {
			my ($dep_task, $num_calls, $parallel, $b) = @{$child_tasks{$task}{$op_id}};
			if ($parallel == 1) {
				$curr_op_time = compute_runtime($dep_task);
			}
			else {
				$curr_op_time = $num_calls * compute_runtime($dep_task);
			}
		}
		else {
			print STDERR "Unrecognized operation type $op\n";
			exit;
		}
		$curr_stage_time = max($curr_op_time, $curr_stage_time);
		debug "curr_op_time = $curr_op_time, curr_stage_time = $curr_stage_time, blocking = $blocking\n";
		if ($blocking) {
			$r += $curr_stage_time;
			$curr_stage_time = 0;
		}
	}
	if ($curr_stage_time != 0) {
		$r += $curr_stage_time;
	}

	debug "Returning runtime of $task as $r\n";

	return $r;
}

sub tco_cost {
	my ($r) = @_;

	#return ${$r}{"price"};
	return ${$r}{"price"} + 3 * 365 * 24 * 0.0001 * ${$r}{"power"};
}

sub io_cap_req {
	my ($s, $read_reqs_ref, $write_reqs_ref, $slowdown) = @_;

	my @read_io_reqs = @{$read_reqs_ref};
	my @write_io_reqs = @{$write_reqs_ref};

	my $seq_reads = 0;
	if ($#write_io_reqs < 0) {
		$seq_reads = 1;
		foreach my $io (@read_io_reqs) {
			my ($op_id, $rate, $record_size, $seq) = @{$io};
			if ($seq != 1) {
				$seq_reads = 0;
				last;
			}
		}
	}

	my $seq_writes = 0;
	if ($#read_io_reqs < 0) {
		$seq_writes = 1;
		foreach my $io (@write_io_reqs) {
			my ($op_id, $rate, $record_size, $seq) = @{$io};
			if ($seq != 1) {
				$seq_writes = 0;
				last;
			}
		}
	}

	my $num_units = 0;
	foreach my $io (@read_io_reqs) {
		my ($op_id, $rate, $record_size, $num_records, $parallel, $seq) = @{$io};
		my $gap = ($seq_reads == 1) ? 0 : ${$s}{"gap_r"};
		my $read_rate = ($record_size / (($record_size / (${$s}{"rate_r"} / $slowdown)) + $gap));
		$num_units += ($rate * $record_size * $num_records / $read_rate);
	}
	foreach my $io (@write_io_reqs) {
		my ($op_id, $rate, $record_size, $num_records, $parallel, $seq) = @{$io};
		my $gap = ($seq_writes == 1) ? 0 : ${$s}{"gap_w"};
		my $write_rate = ($record_size / (($record_size / (${$s}{"rate_w"} / $slowdown)) + $gap));
		$num_units += ($rate * $record_size * $num_records / $write_rate);
	}

	return ${$s}{"cap"} * $num_units;
}

sub set_io_times {
	my ($s, $read_reqs_ref, $write_reqs_ref, $slowdown) = @_;

	my @read_io_reqs = @{$read_reqs_ref};
	my @write_io_reqs = @{$write_reqs_ref};

	my $seq_reads = 0;
	if ($#write_io_reqs < 0) {
		$seq_reads = 1;
		foreach my $io (@read_io_reqs) {
			my ($op_id, $rate, $record_size, $seq) = @{$io};
			if ($seq != 1) {
				$seq_reads = 0;
				last;
			}
		}
	}

	my $seq_writes = 0;
	if ($#read_io_reqs < 0) {
		$seq_writes = 1;
		foreach my $io (@write_io_reqs) {
			my ($op_id, $rate, $record_size, $seq) = @{$io};
			if ($seq != 1) {
				$seq_writes = 0;
				last;
			}
		}
	}

	my $num_units = 0;
	foreach my $io (@read_io_reqs) {
		my ($op_id, $rate, $record_size, $num_records, $parallel, $seq) = @{$io};
		my $gap = ($seq_reads == 1) ? 0 : ${$s}{"gap_r"};
		$io_times{$op_id} = ($parallel == 1 ? 1 : $num_records) * (($record_size / (${$s}{"rate_r"} / $slowdown)) + $gap);
	}
	foreach my $io (@write_io_reqs) {
		my ($op_id, $rate, $record_size, $num_records, $parallel, $seq) = @{$io};
		my $gap = ($seq_writes == 1) ? 0 : ${$s}{"gap_w"};
		$io_times{$op_id} = ($parallel == 1 ? 1 : $num_records) * (($record_size / (${$s}{"rate_w"} / $slowdown)) + $gap);
	}
}

my %dataset_assignments = ();
my %process_assignments = ();
my @machine_assignments = ();

my %compute_reqs = ();
my %storage_reqs_rw = ();
my %storage_reqs_w = ();

sub compute_req_rate {
	my ($task, $phase) = @_;

	if (exists($sla_values{$task})) {
		return $sla_values{$task};
	}
	elsif (exists($parent_tasks{$task})) {
		my $rate = 0;
		foreach my $parent_node (@{$parent_tasks{$task}}) {
			my ($parent, $num_calls) = @{$parent_node};
			if ($phase_hash{$parent} eq $phase) {
				$rate += $num_calls * compute_req_rate($parent, $phase);
			}
		}
		return $rate;
	}
}

sub compute_storage_requirements {
	my ($dataset, $cap, $s_rw, $s_w, $parallelism) = @_;

	#determine IO requirements of dataset
	my %read_io_reqs = ();
	my %write_io_reqs = ();
	my %read_rates = ();
	my %write_rates = ();
	foreach my $phase (keys %phase_sets) {
		$read_rates{$phase} = 0;
		$write_rates{$phase} = 0;
		$read_io_reqs{$phase} = [];
		$write_io_reqs{$phase} = [];
	}

	if (exists($dataset_io{$dataset})) {
	foreach my $op (keys %{$dataset_io{$dataset}}) {
		my ($task, $io_type, $num_records, $record_size, $parallel, $blocking, $seq) = @{$dataset_io{$dataset}{$op}};
		my $phase = $phase_hash{$task};

		my $rate = compute_req_rate($task, $phase);
		if (not(exists($remote_dataset_hash{$dataset}))) {
			$rate /= $parallelism->{$process_hash{$task}};
		}
		if ($io_type eq "R") {
			$read_rates{$phase} += $rate * $num_records * $record_size;
			push @{$read_io_reqs{$phase}}, [$op, $rate, $record_size, $num_records, $parallel, $seq];
		}
		elsif ($io_type eq "W") {
			$write_rates{$phase} += $rate * $num_records * $record_size;
			push @{$write_io_reqs{$phase}}, [$op, $rate, $record_size, $num_records, $parallel, $seq];
		}
		else {
			print STDERR "Unrecognized io_type $io_type\n";
			exit;
		}
	}
	}

	my $max_total_rw_io = 0;
	my $max_total_w_io = 0;
	foreach my $phase (keys %phase_sets) {
		if ($read_rates{$phase} + $write_rates{$phase} > $max_total_rw_io) {
			$max_total_rw_io = $read_rates{$phase} + $write_rates{$phase};
		}
		if ($write_rates{$phase} > $max_total_w_io) {
			$max_total_w_io = $write_rates{$phase};
		}
	}

	my $slowdown = 1;
	if (exists($slowdown_factor{$dataset})) {
		$slowdown = $slowdown_factor{$dataset};
	}
	my $multiplex = 1;
	if (exists($multiplex_factor{$dataset})) {
		$multiplex = $multiplex_factor{$dataset};
	}

	my $max_io_cap_req = 0;
	foreach my $phase (keys %phase_sets) {
		my $io_cap_req = io_cap_req($s_rw, \@{$read_io_reqs{$phase}}, \@{$write_io_reqs{$phase}}, $slowdown);
		if ($io_cap_req > $max_io_cap_req) {
			$max_io_cap_req = $io_cap_req;
		}
		set_io_times($s_rw, \@{$read_io_reqs{$phase}}, \@{$write_io_reqs{$phase}}, $slowdown * $multiplex);
	}
	my $num_units_rw = max($max_io_cap_req, $cap) / ${$s_rw}{"cap"};
	$storage_reqs_rw{$dataset} = [$num_units_rw, $max_total_rw_io / $num_units_rw];

	if ($s_rw ne $s_w) {
		$max_io_cap_req = 0;
		foreach my $phase (keys %phase_sets) {
			my $io_cap_req = io_cap_req($s_w, [], \@{$write_io_reqs{$phase}}, $slowdown * $multiplex);
			if ($io_cap_req > $max_io_cap_req) {
				$max_io_cap_req = $io_cap_req;
			}
			set_io_times($s_w, [], \@{$write_io_reqs{$phase}}, $slowdown);
		}
		my $num_units_w = max($max_io_cap_req, $cap) / ${$s_w}{"cap"};
		$storage_reqs_w{$dataset} = [$num_units_w, $max_total_w_io / $num_units_w];
	}
	else {
		$storage_reqs_w{$dataset} = "";
	}
}

sub compute_task_requirements {
	my ($process, $storage_assignments_rw, $storage_assignments_w, $parallelism) = @_;

	#calling compute_storage_requirements without knowing
	#parallelism to enable setting of IO times
	foreach my $process (keys %process_sets) {
		if (not(exists($parallelism->{$process}))) {
			$parallelism->{$process} = 1;
		}
		foreach my $dataset (@{$process_locality_hash{$process}}) {
			compute_storage_requirements($dataset, $cap_hash{$dataset}, $storage_assignments_rw->{$dataset}, $storage_assignments_w->{$dataset}, $parallelism);
		}
	}

	my %cpu_times = ();
	foreach my $task (@{$process_sets{$process}}) {
		my $phase = $phase_hash{$task};
		$cpu_times{$phase} = 0;
	}
	foreach my $task (@{$process_sets{$process}}) {
		my $phase = $phase_hash{$task};
		foreach my $op_id (keys %{$task_compute{$task}}) {
			my ($compute_time, $blocking) = @{$task_compute{$task}{$op_id}};
			$cpu_times{$phase} += $compute_time;
		}
	}
	my $max_cpu_rate = 0;
	$max_cpu_util{$process} = 0;
	$parallelism->{$process} = 1;
	foreach my $task (@{$process_sets{$process}}) {
		my $phase = $phase_hash{$task};
		my $rate = compute_req_rate($task, $phase);
		if ($rate * $cpu_times{$phase} > $max_cpu_rate) {
			$max_cpu_rate = $rate * $cpu_times{$phase};
		}
		my $runtime = compute_runtime($task);
		if ($cpu_times{$phase} / $runtime > $max_cpu_util{$process}) {
			$max_cpu_util{$process} = $cpu_times{$phase} / $runtime;
		}
		if ($runtime * $rate > $parallelism->{$process}) {
			$parallelism->{$process} = ceil($runtime * $rate);
		}
		debug "task $task: rate = $rate, runtime = $runtime, parallelism = $parallelism->{$process}\n";
	}

	$compute_reqs{$process} = $max_cpu_rate;

	foreach my $process (keys %process_sets) {
		foreach my $dataset (@{$process_locality_hash{$process}}) {
			if ($per_process_copy_hash{$dataset} == 1) {
				compute_storage_requirements($dataset, $cap_hash{$dataset}, $storage_assignments_rw->{$dataset}, $storage_assignments_w->{$dataset}, $parallelism);
			}
			else {
				compute_storage_requirements($dataset, $cap_hash{$dataset} / $parallelism->{$process}, $storage_assignments_rw->{$dataset}, $storage_assignments_w->{$dataset}, $parallelism);
			}
			if ($per_process_copy_hash{$dataset} == 2) {
				${$storage_reqs_rw{$dataset}}[0] = ceil(${$storage_reqs_rw{$dataset}}[0]);
			}
		}
		my $max_memory = 0;
		foreach my $task (@{$process_sets{$process}}) {
			if ($memory_hash{$task} > $max_memory) {
				$max_memory = $memory_hash{$task};
			}
		}
		if ($max_memory > 0) {
			compute_storage_requirements("mem:$process", $max_memory, $storage_assignments_rw->{"mem:$process"}, $storage_assignments_w->{"mem:$process"}, $parallelism);
		}
		else {
			$storage_reqs_rw{"mem:$process"} = [0, 0];
		}
	}
}

my $sys_cost = 0;
sub provision_datasets_tasks {
	my ($storage_assignments_rw, $storage_assignments_w, $parallelism) = @_;

	%dataset_assignments = ();
	%process_assignments = ();
	@machine_assignments = ();
	foreach my $dataset (keys %cap_hash) {
		if (not(exists($remote_dataset_hash{$dataset}))) {
			next;
		}
		debug "Determining storage requirements for remote dataset $dataset\n";

		compute_storage_requirements($dataset, $cap_hash{$dataset}, $storage_assignments_rw->{$dataset}, $storage_assignments_w->{$dataset}, $parallelism);
	}

	foreach my $process (keys %process_sets) {
		debug "Determining requirements for process $process\n";

		compute_task_requirements($process, $storage_assignments_rw, $storage_assignments_w, $parallelism);
	}
}

sub provision_cluster {
	my ($max_cost, $cost_estimate, $num_servers_limit, $storage_assignments_rw, $storage_assignments_w, $parallelism) = @_;
	my @rhs_map = ();
	my %lhs_map = ();

	my $min_cost_estimate = 0;
	foreach my $dataset (keys %cap_hash) {
		if (not(exists($remote_dataset_hash{$dataset}))) {
			next;
		}
		$min_cost_estimate += ${$storage_reqs_rw{$dataset}}[0] * tco_cost($storage_assignments_rw->{$dataset});
		if ($storage_assignments_w->{$dataset} ne $storage_assignments_rw->{$dataset}) {
			$min_cost_estimate += ${$storage_reqs_w{$dataset}}[0] * tco_cost($storage_assignments_w->{$dataset});
		}
	}
	foreach my $process (keys %process_sets) {
		$min_cost_estimate += ceil($parallelism->{$process} * $max_cpu_util{$process}) * tco_cost($compute_res[0]);
		foreach my $dataset (@{$process_locality_hash{$process}}) {
			$min_cost_estimate += ceil(${$storage_reqs_rw{$dataset}}[0] * $parallelism->{$process}) * tco_cost($storage_assignments_rw->{$dataset});
			if ($storage_assignments_w->{$dataset} ne $storage_assignments_rw->{$dataset}) {
				$min_cost_estimate += ceil(${$storage_reqs_w{$dataset}}[0] * $parallelism->{$process}) * tco_cost($storage_assignments_w->{$dataset});
			}
		}
	}
	if ($cost_estimate == 1) {
		return $min_cost_estimate;
	}
	if ($min_cost_estimate > $max_cost) {
		debug "Cost estimate $min_cost_estimate greater than current system cost $max_cost\n";
		#return $INFINITY;
	}

	my %min_units_limit = ();
	my %min_io_limit = ();
	my $min_cpu_limit = $INFINITY;
	my $min_nw_limit = $INFINITY;
	foreach my $server (@servers_res) {
		foreach my $bus (keys %{${$server}{"b"}}) {
			my ($units_limit, $io_limit) = @{${$server}{"b"}{$bus}};
			if (not(exists($min_units_limit{$bus})) or $units_limit < $min_units_limit{$bus}) {
				$min_units_limit{$bus} = $units_limit;
			}
			if (not(exists($min_io_limit{$bus})) or $io_limit < $min_io_limit{$bus}) {
				$min_io_limit{$bus} = $io_limit;
			}
		}
		if (${$server}{"c"}{${$compute_res[0]}{"name"}} < $min_cpu_limit) {
			$min_cpu_limit = ${$server}{"c"}{${$compute_res[0]}{"name"}};
		}
		if (${$server}{"nwlimit"} < $min_nw_limit) {
			$min_nw_limit = ${$server}{"nwlimit"};
		}
	}

	my $num_servers_s = 0;
	my %num_servers = ();
	foreach my $dataset (keys %cap_hash) {
		my $s_rw = $storage_assignments_rw->{$dataset};
		my $s_w = $storage_assignments_w->{$dataset};
		if (exists($remote_dataset_hash{$dataset})) {
			my $num_servers_s_1 = ${$storage_reqs_rw{$dataset}}[0] / $min_units_limit{${$s_rw}{"bus"}};
			my $num_servers_s_2 = ${$storage_reqs_rw{$dataset}}[0] * ${$storage_reqs_rw{$dataset}}[1] / $min_io_limit{${$s_rw}{"bus"}};
			my $num_servers_s_3 = ${$storage_reqs_rw{$dataset}}[0] * ${$storage_reqs_rw{$dataset}}[1] / $min_nw_limit;
			if ($remote_dataset_hash{$dataset} == 1) {
				$num_servers_s += max3(ceil($num_servers_s_1), ceil($num_servers_s_2), ceil($num_servers_s_3));
			}
			else {
				$num_servers{$dataset} = max3(ceil($num_servers_s_1), ceil($num_servers_s_2), ceil($num_servers_s_3));
			}
			if ($storage_reqs_w{$dataset} ne "") {
				my $num_servers_s_1 = ${$storage_reqs_w{$dataset}}[0] / $min_units_limit{${$s_w}{"bus"}};
				my $num_servers_s_2 = ${$storage_reqs_w{$dataset}}[0] * ${$storage_reqs_w{$dataset}}[1] / $min_io_limit{${$s_w}{"bus"}};
				my $num_servers_s_3 = ${$storage_reqs_w{$dataset}}[0] * ${$storage_reqs_w{$dataset}}[1] / $min_nw_limit;
				if ($remote_dataset_hash{$dataset} == 1) {
					$num_servers_s += max3(ceil($num_servers_s_1), ceil($num_servers_s_2), ceil($num_servers_s_3));
				}
				else {
					$num_servers{$dataset} += max3(ceil($num_servers_s_1), ceil($num_servers_s_2), ceil($num_servers_s_3));
				}
			}
			if ($remote_dataset_hash{$dataset} != 1) {
				debug "num_servers for $dataset = $num_servers{$dataset}\n";
			}
		}
	}
	my $num_servers_c = 0;
	my $num_servers_s_local = 0;
	foreach my $process (keys %process_sets) {
		my $num_servers_c += ceil($parallelism->{$process} / int($min_cpu_limit / $max_cpu_util{$process}));
		if (exists($process_locality_hash{$process})) {
			foreach my $dataset (@{$process_locality_hash{$process}}) {
				my $s_rw = $storage_assignments_rw->{$dataset};
				my $s_w = $storage_assignments_w->{$dataset};
				my $num_servers_s_1 = ceil($parallelism->{$process} / int($min_units_limit{${$s_rw}{"bus"}} / ${$storage_reqs_rw{$dataset}}[0]));
				my $num_servers_s_2 = ceil($parallelism->{$process} / int($min_io_limit{${$s_rw}{"bus"}} / ${$storage_reqs_rw{$dataset}}[1]));
				$num_servers_s_local += max($num_servers_s_1, $num_servers_s_2);
				if ($storage_reqs_w{$dataset} ne "") {
					my $num_servers_s_1 = ceil($parallelism->{$process} / int($min_units_limit{${$s_w}{"bus"}} / ${$storage_reqs_w{$dataset}}[0]));
					my $num_servers_s_2 = ceil($parallelism->{$process} / int($min_io_limit{${$s_w}{"bus"}} / ${$storage_reqs_w{$dataset}}[1]));
					$num_servers_s_local += max($num_servers_s_1, $num_servers_s_2);
				}
			}
		}
	}
	debug "num_servers for storage = $num_servers_s, num_servers for compute = $num_servers_c, num_servers for local storage = $num_servers_s_local\n";
	my $max_num_servers_1 = max3($num_servers_s, $num_servers_c, $num_servers_s_local);
	my $max_num_servers = min($max_num_servers_1, $num_servers_limit);

	my $num_constraints = 0;
	my @datasets = ();
	my @processes = keys %process_sets;
	my @remote_datasets = ();
	foreach my $dataset (keys %cap_hash) {
		if (exists($remote_dataset_hash{$dataset})) {
			if ($remote_dataset_hash{$dataset} == 1) {
				push @datasets, $dataset;
			}
			else {
				push @remote_datasets, $dataset;
			}
		}
	}

	#initialize variables
	for (my $i = 0; $i < $max_num_servers; ++$i) {
		for (my $j = 0; $j < @servers_res; ++$j) {
			$lhs_map{"server_type_$i\_$j"} = [];
		}
		for (my $j = 0; $j < @datasets; ++$j) {
			my $dataset = $datasets[$j];
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			$lhs_map{"num_units_$i\_$j\_$s_rw_name"} = [];
			if ($s_w ne $s_rw) {
				$lhs_map{"num_units_$i\_$j\_$s_w_name"} = [];
			}
		}
		for (my $j = 0; $j < @processes; ++$j) {
			$lhs_map{"parallelism_$i\_$j"} = [];
		}
	}
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
			for (my $j = 0; $j < @servers_res; ++$j) {
				$lhs_map{"server_type_$n\_$i\_$j"} = [];
			}
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			$lhs_map{"num_units_$i\_r$n\_$s_rw_name"} = [];
			if ($s_w ne $s_rw) {
				$lhs_map{"num_units_$i\_r$n\_$s_w_name"} = [];
			}
		}
	}

	#for each type of bus, enforce number of units limit
	for (my $k = 0; $k < @bus_res; ++$k) {
		for (my $i = 0; $i < $max_num_servers; ++$i) {
			++$num_constraints;
			for (my $j = 0; $j < @servers_res; ++$j) {
				if (exists(${$servers_res[$j]}{"b"}{$bus_res[$k]})) {
					push @{$lhs_map{"server_type_$i\_$j"}}, [$num_constraints, -${${$servers_res[$j]}{"b"}{$bus_res[$k]}}[0]];
				}
			}
			for (my $j = 0; $j < @datasets; ++$j) {
				my $dataset = $datasets[$j];
				my $s_rw = $storage_assignments_rw->{$dataset};
				my $s_rw_name = ${$s_rw}{"name"};
				my $s_w = $storage_assignments_w->{$dataset};
				my $s_w_name = ${$s_w}{"name"};
				if (${$s_rw}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_$j\_$s_rw_name"}}, [$num_constraints, 1];
				}
				if ($s_rw ne $s_w and ${$s_w}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_$j\_$s_w_name"}}, [$num_constraints, 1];
				}
			}
			for (my $j = 0; $j < @processes; ++$j) {
				my $process = $processes[$j];
				my $mem = $storage_assignments_rw->{"mem:$process"};
				my $coeff = 0;
				if (${$mem}{"bus"} eq $bus_res[$k]) {
					$coeff += ${$storage_reqs_rw{"mem:$process"}}[0];
				}
				foreach my $dataset (@{$process_locality_hash{$process}}) {
					my $s_rw = $storage_assignments_rw->{$dataset};
					my $s_rw_name = ${$s_rw}{"name"};
					my $s_w = $storage_assignments_w->{$dataset};
					my $s_w_name = ${$s_w}{"name"};
					if (${$s_rw}{"bus"} eq $bus_res[$k]) {
						$coeff += ${$storage_reqs_rw{$dataset}}[0];
					}
					if ($s_rw ne $s_w and ${$s_w}{"bus"} eq $bus_res[$k]) {
						$coeff += ${$storage_reqs_w{$dataset}}[0];
					}
				}
				if ($coeff != 0) {
					push @{$lhs_map{"parallelism_$i\_$j"}}, [$num_constraints, $coeff];
				}
			}
			push @rhs_map, 0; 
		}
		for (my $n = 0; $n < @remote_datasets; ++$n) {
			my $dataset = $remote_datasets[$n];
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
				++$num_constraints;
				for (my $j = 0; $j < @servers_res; ++$j) {
					if (exists(${$servers_res[$j]}{"b"}{$bus_res[$k]})) {
						push @{$lhs_map{"server_type_$n\_$i\_$j"}}, [$num_constraints, -${${$servers_res[$j]}{"b"}{$bus_res[$k]}}[0]];
					}
				}
				if (${$s_rw}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_r$n\_$s_rw_name"}}, [$num_constraints, 1];
				}
				if ($s_rw ne $s_w and ${$s_w}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_r$n\_$s_w_name"}}, [$num_constraints, 1];
				}
				push @rhs_map, 0;
			}
		}
	}

	#for each type of bus, enforce IO BW limit
	for (my $k = 0; $k < @bus_res; ++$k) {
		for (my $i = 0; $i < $max_num_servers; ++$i) {
			++$num_constraints;
			for (my $j = 0; $j < @servers_res; ++$j) {
				if (exists(${$servers_res[$j]}{"b"}{$bus_res[$k]})) {
					push @{$lhs_map{"server_type_$i\_$j"}}, [$num_constraints, -${${$servers_res[$j]}{"b"}{$bus_res[$k]}}[1]];
				}
			}
			for (my $j = 0; $j < @datasets; ++$j) {
				my $dataset = $datasets[$j];
				my $s_rw = $storage_assignments_rw->{$dataset};
				my $s_rw_name = ${$s_rw}{"name"};
				my $s_w = $storage_assignments_w->{$dataset};
				my $s_w_name = ${$s_w}{"name"};
				if (${$s_rw}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_$j\_$s_rw_name"}}, [$num_constraints, ${$storage_reqs_rw{$dataset}}[1]];
				}
				if ($s_rw ne $s_w and ${$s_w}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_$j\_$s_w_name"}}, [$num_constraints, ${$storage_reqs_w{$dataset}}[1]];
				}
			}
			for (my $j = 0; $j < @processes; ++$j) {
				my $process = $processes[$j];
				my $mem = $storage_assignments_rw->{"mem:$process"};
				my $coeff = 0;
				if (${$mem}{"bus"} eq $bus_res[$k]) {
					$coeff += ${$storage_reqs_rw{"mem:$process"}}[1];
				}
				foreach my $dataset (@{$process_locality_hash{$process}}) {
					my $s_rw = $storage_assignments_rw->{$dataset};
					my $s_rw_name = ${$s_rw}{"name"};
					my $s_w = $storage_assignments_w->{$dataset};
					my $s_w_name = ${$s_w}{"name"};
					if (${$s_rw}{"bus"} eq $bus_res[$k]) {
						$coeff += ${$storage_reqs_rw{$dataset}}[1];
					}
					if ($s_rw ne $s_w and ${$s_w}{"bus"} eq $bus_res[$k]) {
						$coeff += ${$storage_reqs_w{$dataset}}[1];
					}
				}
				if ($coeff != 0) {
					push @{$lhs_map{"parallelism_$i\_$j"}}, [$num_constraints, $coeff];
				}
			}
			push @rhs_map, 0; 
		}
		for (my $n = 0; $n < @remote_datasets; ++$n) {
			my $dataset = $remote_datasets[$n];
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
				++$num_constraints;
				for (my $j = 0; $j < @servers_res; ++$j) {
					if (exists(${$servers_res[$j]}{"b"}{$bus_res[$k]})) {
						push @{$lhs_map{"server_type_$n\_$i\_$j"}}, [$num_constraints, -${${$servers_res[$j]}{"b"}{$bus_res[$k]}}[1]];
					}
				}
				if (${$s_rw}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_r$n\_$s_rw_name"}}, [$num_constraints, ${$storage_reqs_rw{$dataset}}[1]];
				}
				if ($s_rw ne $s_w and ${$s_w}{"bus"} eq $bus_res[$k]) {
					push @{$lhs_map{"num_units_$i\_r$n\_$s_w_name"}}, [$num_constraints, ${$storage_reqs_w{$dataset}}[1]];
				}
				push @rhs_map, 0;
			}
		}
	}

	#enforce network BW limits
	for (my $i = 0; $i < $max_num_servers; ++$i) {
		++$num_constraints;
		for (my $j = 0; $j < @servers_res; ++$j) {
			push @{$lhs_map{"server_type_$i\_$j"}}, [$num_constraints, -${$servers_res[$j]}{"nwlimit"}];
		}
		for (my $j = 0; $j < @datasets; ++$j) {
			my $dataset = $datasets[$j];
			if (not(exists($remote_dataset_hash{$dataset}))) {
				next;
			}
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			push @{$lhs_map{"num_units_$i\_$j\_$s_rw_name"}}, [$num_constraints, ${$storage_reqs_rw{$dataset}}[1]];
			if ($s_rw ne $s_w) {
				push @{$lhs_map{"num_units_$i\_$j\_$s_w_name"}}, [$num_constraints, ${$storage_reqs_w{$dataset}}[1]];
			}
		}
		push @rhs_map, 0; 
	}
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		my $s_rw = $storage_assignments_rw->{$dataset};
		my $s_rw_name = ${$s_rw}{"name"};
		my $s_w = $storage_assignments_w->{$dataset};
		my $s_w_name = ${$s_w}{"name"};
		for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
			++$num_constraints;
			for (my $j = 0; $j < @servers_res; ++$j) {
				push @{$lhs_map{"server_type_$n\_$i\_$j"}}, [$num_constraints, -${$servers_res[$j]}{"nwlimit"}];
			}
			push @{$lhs_map{"num_units_$i\_r$n\_$s_rw_name"}}, [$num_constraints, ${$storage_reqs_rw{$dataset}}[1]];
			if ($s_rw ne $s_w) {
				push @{$lhs_map{"num_units_$i\_r$n\_$s_w_name"}}, [$num_constraints, ${$storage_reqs_w{$dataset}}[1]];
			}
			push @rhs_map, 0;
		}
	}

	#enforce CPU limits
	for (my $i = 0; $i < $max_num_servers; ++$i) {
		++$num_constraints;
		for (my $j = 0; $j < @servers_res; ++$j) {
			push @{$lhs_map{"server_type_$i\_$j"}}, [$num_constraints, -${${$servers_res[$j]}{"c"}{${$compute_res[0]}{"name"}}}[0]];
		}
		for (my $j = 0; $j < @processes; ++$j) {
			my $process = $processes[$j];
			push @{$lhs_map{"parallelism_$i\_$j"}}, [$num_constraints, $max_cpu_util{$process}];
		}
		push @rhs_map, 0; 
	}

	#choose at most one type of server per machine
	for (my $i = 0; $i < $max_num_servers; ++$i) {
		++$num_constraints;
		for (my $j = 0; $j < @servers_res; ++$j) {
			push @{$lhs_map{"server_type_$i\_$j"}}, [$num_constraints, 1];
		}
		push @rhs_map, 1;
	}
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
			++$num_constraints;
			for (my $j = 0; $j < @servers_res; ++$j) {
				push @{$lhs_map{"server_type_$n\_$i\_$j"}}, [$num_constraints, 1];
			}
			push @rhs_map, 1;
		}
	}

	#never choose "(i + 1)^th server, but not i^th server"
	for (my $i = 0; $i < $max_num_servers - 1; ++$i) {
		++$num_constraints;
		for (my $j = 0; $j < @servers_res; ++$j) {
			push @{$lhs_map{"server_type_$i\_$j"}}, [$num_constraints, -1];
		}
		my $k = $i + 1;
		for (my $j = 0; $j < @servers_res; ++$j) {
			push @{$lhs_map{"server_type_$k\_$j"}}, [$num_constraints, 1];
		}
		push @rhs_map, 0;
	}
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		for (my $i = 0; $i < $num_servers{$dataset} - 1; ++$i) {
			++$num_constraints;
			for (my $j = 0; $j < @servers_res; ++$j) {
				push @{$lhs_map{"server_type_$n\_$i\_$j"}}, [$num_constraints, -1];
			}
			my $k = $i + 1;
			for (my $j = 0; $j < @servers_res; ++$j) {
				push @{$lhs_map{"server_type_$n\_$k\_$j"}}, [$num_constraints, 1];
			}
			push @rhs_map, 0;
		}
	}

	++$num_constraints;
	foreach my $var (keys %lhs_map) {
		if ($var =~ m/^server_type_(\d+)_(\d+)$/) {
			my $server_index = $2;
			my $cost = tco_cost($servers_res[$server_index]);
			push @{$lhs_map{$var}}, [$num_constraints, $cost];
		}
		if ($var =~ m/^server_type_(\d+)_(\d+)_(\d+)$/) {
			my $server_index = $3;
			my $cost = tco_cost($servers_res[$server_index]);
			push @{$lhs_map{$var}}, [$num_constraints, $cost];
		}
	}
	push @rhs_map, $max_cost;
	my $num_inequality_constraints = $num_constraints;

	#ensure num storage units add up
	for (my $j = 0; $j < @datasets; ++$j) {
		my $dataset = $datasets[$j];
		my $s_rw = $storage_assignments_rw->{$dataset};
		my $s_rw_name = ${$s_rw}{"name"};
		my $s_w = $storage_assignments_w->{$dataset};
		my $s_w_name = ${$s_w}{"name"};
		++$num_constraints;
		for (my $i = 0; $i < $max_num_servers; ++$i) {
			push @{$lhs_map{"num_units_$i\_$j\_$s_rw_name"}}, [$num_constraints, 1];
		}
		push @rhs_map, ceil(${$storage_reqs_rw{$dataset}}[0]);
		if ($s_w ne $s_rw) {
			++$num_constraints;
			for (my $i = 0; $i < $max_num_servers; ++$i) {
				push @{$lhs_map{"num_units_$i\_$j\_$s_w_name"}}, [$num_constraints, 1];
			}
			push @rhs_map, ceil(${$storage_reqs_w{$dataset}}[0]);
		}
	}
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		my $s_rw = $storage_assignments_rw->{$dataset};
		my $s_rw_name = ${$s_rw}{"name"};
		my $s_w = $storage_assignments_w->{$dataset};
		my $s_w_name = ${$s_w}{"name"};
		++$num_constraints;
		for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
			push @{$lhs_map{"num_units_$i\_r$n\_$s_rw_name"}}, [$num_constraints, 1];
		}
		push @rhs_map, ceil(${$storage_reqs_rw{$dataset}}[0]);
		if ($s_w ne $s_rw) {
			++$num_constraints;
			for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
				push @{$lhs_map{"num_units_$i\_r$n\_$s_w_name"}}, [$num_constraints, 1];
			}
			push @rhs_map, ceil(${$storage_reqs_w{$dataset}}[0]);
		}
	}

	#ensure parallelism adds up
	for (my $j = 0; $j < @processes; ++$j) {
		++$num_constraints;
		my $process = $processes[$j];
		for (my $i = 0; $i < $max_num_servers; ++$i) {
			push @{$lhs_map{"parallelism_$i\_$j"}}, [$num_constraints, 1];
		}
		push @rhs_map, $parallelism->{$process};
	}

	open MIP, ">tmp_ip_formulation.txt";
	print MIP "NAME\tSCC\n";
	print MIP "ROWS\n";
	print MIP "\tN\tcost\n";
	for (my $i = 0; $i < $num_constraints; ++$i) {
		if ($i < $num_inequality_constraints) {
			print MIP "\tL\tR$i\n";
		}
		else {
			print MIP "\tE\tR$i\n";
		}
	}
	print MIP "COLUMNS\n";
	print MIP "\tMARK0001\t'MARKER'\t'INTORG'\n";
	foreach my $var (keys %lhs_map) {
		foreach my $clause (@{$lhs_map{$var}}) {
			my ($r_num, $coeff) = @{$clause};
			--$r_num;
			print MIP "\t$var\tR$r_num\t$coeff\n";
		}
		if ($var =~ m/^server_type_(\d+)_(\d+)$/) {
			my $server_index = $2;
			my $cost = tco_cost($servers_res[$server_index]);
			print MIP "\t$var\tcost\t$cost\n";
		}
		if ($var =~ m/^server_type_(\d+)_(\d+)_(\d+)$/) {
			my $server_index = $3;
			my $cost = tco_cost($servers_res[$server_index]);
			print MIP "\t$var\tcost\t$cost\n";
		}
	}
	print MIP "\tMARK0002\t'MARKER'\t'INTEND'\n";
	print MIP "RHS\n";
	for (my $i = 0; $i < $num_constraints; ++$i) {
		print MIP "\tRHS1\tR$i\t$rhs_map[$i]\n";
	}
	print MIP "BOUNDS\n";
	for (my $i = 0; $i < $max_num_servers; ++$i) {
		for (my $j = 0; $j < @servers_res; ++$j) {
			print MIP " UP BND1\tserver_type_$i\_$j\t1\n";
			print MIP " LO BND1\tserver_type_$i\_$j\t0\n";
		}
		for (my $j = 0; $j < @datasets; ++$j) {
			my $dataset = $datasets[$j];
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			my $num_units_rw = ceil(${$storage_reqs_rw{$dataset}}[0]);
			print MIP " UP BND1\tnum_units_$i\_$j\_$s_rw_name\t$num_units_rw\n";
			print MIP " LO BND1\tnum_units_$i\_$j\_$s_rw_name\t0\n";
			if ($s_w ne $s_rw) {
				my $num_units_w = ceil(${$storage_reqs_w{$dataset}}[0]);
				print MIP " UP BND1\tnum_units_$i\_$j\_$s_w_name\t$num_units_w\n";
				print MIP " LO BND1\tnum_units_$i\_$j\_$s_w_name\t0\n";
			}
		}
		for (my $j = 0; $j < @processes; ++$j) {
			my $process = $processes[$j];
			print MIP " UP BND1\tparallelism_$i\_$j\t$parallelism->{$process}\n";
			print MIP " LO BND1\tparallelism_$i\_$j\t0\n";
		}
	}
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		my $s_rw = $storage_assignments_rw->{$dataset};
		my $s_rw_name = ${$s_rw}{"name"};
		my $s_w = $storage_assignments_w->{$dataset};
		my $s_w_name = ${$s_w}{"name"};
		my $num_units_rw = ceil(${$storage_reqs_rw{$dataset}}[0]);
		for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
			for (my $j = 0; $j < @servers_res; ++$j) {
				print MIP " UP BND1\tserver_type_$n\_$i\_$j\t1\n";
				print MIP " LO BND1\tserver_type_$n\_$i\_$j\t0\n";
			}
			print MIP " UP BND1\tnum_units_$i\_r$n\_$s_rw_name\t$num_units_rw\n";
			print MIP " LO BND1\tnum_units_$i\_r$n\_$s_rw_name\t0\n";
			if ($s_w ne $s_rw) {
				my $num_units_w = ceil(${$storage_reqs_w{$dataset}}[0]);
				print MIP " UP BND1\tnum_units_$i\_r$n\_$s_w_name\t$num_units_w\n";
				print MIP " LO BND1\tnum_units_$i\_r$n\_$s_w_name\t0\n";
			}
		}
	}
	print MIP "ENDATA\n";
	close MIP;

	my $start = time;
	system "./glpsol --tmlim 10 --pcost -o tmp_ip_output tmp_ip_formulation.txt > /dev/null";
	my $end = time;
	my $runtime = $end - $start;
	debug "Time to run LP = $runtime\n";

	open OUT, "tmp_ip_output";
	while (<OUT>) {
		chomp;
		if (m/INTEGER UNDEFINED/ or m/INTEGER EMPTY/) {
			if ($max_num_servers < $num_servers_limit) {
				return $INFINITY;
			}
			else {
				return -1;
			}
		}
		if (m/Column name/) {
			my $line = <OUT>;
			last;
		}
	}

	my %assigned_val = ();
	while (<OUT>) {
		chomp;
		if ($_ eq "") {
			last;
		}
		my @f = split " ";
		my $var = $f[1];
		my $line = <OUT>;
		chomp $line;
		@f = split " ", $line;
		my $val = $f[1];

		$assigned_val{$var} = $val;
	}
	close OUT;

	my $sys_cost = 0;
	for (my $n = 0; $n < @remote_datasets; ++$n) {
		my $dataset = $remote_datasets[$n];
		my $s_rw = $storage_assignments_rw->{$dataset};
		my $s_rw_name = ${$s_rw}{"name"};
		my $s_w = $storage_assignments_w->{$dataset};
		my $s_w_name = ${$s_w}{"name"};
		my $num_units_rw = ceil(${$storage_reqs_rw{$dataset}}[0]);
		for (my $i = 0; $i < $num_servers{$dataset}; ++$i) {
			my $server_found = 0;
			for (my $j = 0; $j < @servers_res; ++$j) {
				if ($assigned_val{"server_type_$n\_$i\_$j"} == 1) {
					if ($silent_mode != 1) {
					printf "Server $i (%s):\n", ${$servers_res[$j]}{"name"};
					}
					$sys_cost += tco_cost($servers_res[$j]);
					$server_found = 1;
					last;
				}
			}
			if ($server_found == 0) {
				last;
			}
			if ($assigned_val{"num_units_$i\_r$n\_$s_rw_name"} > 0) {
				if ($silent_mode != 1) {
				printf "\t%d units of $s_rw_name for dataset $dataset\n", $assigned_val{"num_units_$i\_r$n\_$s_rw_name"};
				}
				$sys_cost += $assigned_val{"num_units_$i\_r$n\_$s_rw_name"} * tco_cost($storage_assignments_rw->{$dataset});
			}
			if ($s_w ne $s_rw) {
				if ($assigned_val{"num_units_$i\_r$n\_$s_w_name"} > 0) {
					if ($silent_mode != 1) {
					printf "\t%d units of $s_w_name for dataset $dataset\n", $assigned_val{"num_units_$i\_r$n\_$s_w_name"};
					}
					$sys_cost += $assigned_val{"num_units_$i\_r$n\_$s_w_name"} * tco_cost($storage_assignments_w->{$dataset});
				}
			}
			#every machine should have at least one CPU
			if ($silent_mode != 1) {
			print "\t1 CPU\n";
			}
			$sys_cost += tco_cost($compute_res[0]);
		}
	}
	for (my $i = 0; $i < $max_num_servers; ++$i) {
		my $server_found = 0;
		for (my $j = 0; $j < @servers_res; ++$j) {
			if ($assigned_val{"server_type_$i\_$j"} == 1) {
				if ($silent_mode != 1) {
				printf "Server $i (%s):\n", ${$servers_res[$j]}{"name"};
				}
				$sys_cost += tco_cost($servers_res[$j]);
				$server_found = 1;
				last;
			}
		}
		if ($server_found == 0) {
			last;
		}

		for (my $j = 0; $j < @datasets; ++$j) {
			my $dataset = $datasets[$j];
			my $s_rw = $storage_assignments_rw->{$dataset};
			my $s_rw_name = ${$s_rw}{"name"};
			my $s_w = $storage_assignments_w->{$dataset};
			my $s_w_name = ${$s_w}{"name"};
			if ($assigned_val{"num_units_$i\_$j\_$s_rw_name"} > 0) {
				if ($silent_mode != 1) {
				printf "\t%d units of $s_rw_name for dataset $dataset\n", $assigned_val{"num_units_$i\_$j\_$s_rw_name"};
				}
				$sys_cost += $assigned_val{"num_units_$i\_$j\_$s_rw_name"} * tco_cost($storage_assignments_rw->{$dataset});
			}
			if ($s_w ne $s_rw) {
				if ($assigned_val{"num_units_$i\_$j\_$s_w_name"} > 0) {
					if ($silent_mode != 1) {
					printf "\t%d units of $s_w_name for dataset $dataset\n", $assigned_val{"num_units_$i\_$j\_$s_w_name"};
					}
					$sys_cost += $assigned_val{"num_units_$i\_$j\_$s_w_name"} * tco_cost($storage_assignments_w->{$dataset});
				}
			}
		}
		for (my $j = 0; $j < @processes; ++$j) {
			my $process = $processes[$j];
			if ($assigned_val{"parallelism_$i\_$j"} > 0) {
				my $num_cpus = ceil($max_cpu_util{$process} * $assigned_val{"parallelism_$i\_$j"});
				$sys_cost += $num_cpus * tco_cost($compute_res[0]);
				if ($silent_mode != 1) {
				printf "\t%d CPUs for %d instances of $process\n", $num_cpus, $assigned_val{"parallelism_$i\_$j"};
				}
				foreach my $dataset (@{$process_locality_hash{$process}}) {
					my $s_rw = $storage_assignments_rw->{$dataset};
					my $s_rw_name = ${$s_rw}{"name"};
					my $s_w = $storage_assignments_w->{$dataset};
					my $s_w_name = ${$s_w}{"name"};
					if ($silent_mode != 1) {
					printf "\t%d units of $s_rw_name for $dataset\n", ceil(${$storage_reqs_rw{$dataset}}[0] * $assigned_val{"parallelism_$i\_$j"});
					}
					$sys_cost += ceil(${$storage_reqs_rw{$dataset}}[0] * $assigned_val{"parallelism_$i\_$j"}) * tco_cost($storage_assignments_rw->{$dataset});
					if ($s_w ne $s_rw) {
						if ($silent_mode != 1) {
						printf "\t%d units of $s_w_name for $dataset\n", ceil(${$storage_reqs_w{$dataset}}[0] * $assigned_val{"parallelism_$i\_$j"});
						}
						$sys_cost += ceil(${$storage_reqs_w{$dataset}}[0] * $assigned_val{"parallelism_$i\_$j"}) * tco_cost($storage_assignments_w->{$dataset});
					}
				}
			}
		}
	}

	if ($sys_cost == 0) {
		print STDERR "System cost shouldn't be 0 if a feasible solution was found\n";
		exit;
	}
	return $sys_cost;
}

sub provision_cluster_wrapper {
	my ($max_cost, $cost_estimate, $storage_assignments_rw, $storage_assignments_w, $parallelism) = @_;

	my $num_servers_limit = 1;
	while (1) {
		my $sys_cost = provision_cluster($max_cost, $cost_estimate, $num_servers_limit, $storage_assignments_rw, $storage_assignments_w, $parallelism);
		if ($sys_cost > 0) {
			return $sys_cost;
		}
		++$num_servers_limit;
	}
}

sub get_storage {
	my ($persistence, $index) = @_;

	my $count = 0;
	foreach my $s (@storage_res) {
		if (($persistence eq "1" or $persistence eq "2") and ${$s}{"persistence"} eq "0") {
			#Can't use volatile storage for non-volatile data
			next;
		}
		if ($persistence eq "0" and ${$s}{"persistence"} eq "1") {
			#Can't use non-volatile storage for volatile data
			next;
		}
		#printf "Considering $count: (%s, %s)\n", ${$s}{"name"}, ${$s}{"name"};
		if ($count == $index) {
			return ($s, $s);
		}
		++$count;
	}

	if ($persistence eq "1") {
	foreach my $s_w (@storage_res) { #unit to use for writes
		if (${$s_w}{"persistence"} ne "1") {
			#medium used for writes needs to be persistent
			next;
		}
		foreach my $s_rw (@storage_res) { #unit to use for reads
			if ($s_w eq $s_rw) {
				next;
			}
			if (${$s_rw}{"persistence"} eq "1") {
				#medium used for reads needs to be volatile
				next;
			}
			#printf "Considering $count: (%s, %s)\n", ${$s_w}{"name"}, ${$s_rw}{"name"};
			if ($count == $index) {
				return ($s_w, $s_rw);
			}
			++$count;
		}
	}
	}

	return ("", "");
}

#configuration state
my %storage_assignments_rw;
my %storage_assignments_w;
my %parallelism = ();

sub run_model {
	%storage_assignments_rw = ();
	%storage_assignments_w = ();

	my @num_storage_options = ();
	for (my $i = 0; $i <= 2; ++$i) {
		push @num_storage_options, 0;
		while (1) {
			my ($s_w, $s_rw) = get_storage($i, $num_storage_options[$i]);
			if ($s_w eq "") {
				last;
			}
			++$num_storage_options[$i];
		}
	}

	foreach my $dataset (keys %cap_hash) {
		my $index = int(rand($num_storage_options[$persistence_hash{$dataset}]));
		($storage_assignments_w{$dataset}, $storage_assignments_rw{$dataset}) = get_storage($persistence_hash{$dataset}, $index);
	}
	foreach my $process (keys %process_sets) {
		($storage_assignments_rw{"mem:$process"}, $storage_assignments_w{"mem:$process"}) = get_storage("0", 0);
	}

	provision_datasets_tasks(\%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
	my $sys_cost = provision_cluster_wrapper($INFINITY, 1, \%storage_assignments_rw, \%storage_assignments_w, \%parallelism);

	if ($SELECT_INIT_CONFIG) {
	while (1) {
		if ($silent_mode != 1) {
			print "\nMoving on to configuration:\n";
			display_assignment(\%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
			print "System cost = $sys_cost\n";
		}

		my $min_cost = $sys_cost;
		my @min_cost_perturbation = ();
		my %tmp_storage_assignments_w = ();
		my %tmp_storage_assignments_rw = ();

		foreach my $dataset (keys %cap_hash) {
			my $index = -1;
			while (1) {
				foreach my $d (keys %storage_assignments_w) {
					$tmp_storage_assignments_w{$d} = $storage_assignments_w{$d};
					$tmp_storage_assignments_rw{$d} = $storage_assignments_rw{$d};
				}
				++$index;
				debug "\nAssigning $dataset to storage option $index\n";
				($tmp_storage_assignments_w{$dataset}, $tmp_storage_assignments_rw{$dataset}) = get_storage($persistence_hash{$dataset}, $index);
				if ($tmp_storage_assignments_w{$dataset} eq "") {
					last;
				}
				if ($tmp_storage_assignments_w{$dataset} eq $storage_assignments_w{$dataset} and $tmp_storage_assignments_rw{$dataset} eq $storage_assignments_rw{$dataset}) {
					next;
				}

				provision_datasets_tasks(\%tmp_storage_assignments_rw, \%tmp_storage_assignments_w, \%parallelism);
				debug "\nTrying out configuration:\n";
				if ($DEBUG) {
				display_assignment(\%tmp_storage_assignments_rw, \%tmp_storage_assignments_w, \%parallelism);
				}
				my $cost = provision_cluster_wrapper($min_cost, 1, \%tmp_storage_assignments_rw, \%tmp_storage_assignments_w, \%parallelism);
				debug "Cost = $cost\n";
				if ($cost < $min_cost) {
					debug "Updating min_cost from $min_cost to $cost\n";
					@min_cost_perturbation = ($dataset, $tmp_storage_assignments_w{$dataset}, $tmp_storage_assignments_rw{$dataset});
					$min_cost = $cost;
				}
			}
		}

		if ($#min_cost_perturbation < 0) {
			last;
		}
		$sys_cost = $min_cost;
		my $dataset = $min_cost_perturbation[0];
		$storage_assignments_w{$dataset} = $min_cost_perturbation[1];
		$storage_assignments_rw{$dataset} = $min_cost_perturbation[2];
	}

	provision_datasets_tasks(\%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
	$sys_cost = provision_cluster_wrapper($INFINITY, 0, \%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
	}

	while (1) {
		if ($silent_mode != 1) {
			print "\nMoving on to configuration:\n";
			display_assignment(\%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
			print "System cost = $sys_cost\n";
		}

		my $min_cost = $sys_cost;
		my @min_cost_perturbation = ();
		my %tmp_storage_assignments_w = ();
		my %tmp_storage_assignments_rw = ();

		foreach my $dataset (keys %cap_hash) {
			my $index = -1;
			while (1) {
				foreach my $d (keys %storage_assignments_w) {
					$tmp_storage_assignments_w{$d} = $storage_assignments_w{$d};
					$tmp_storage_assignments_rw{$d} = $storage_assignments_rw{$d};
				}
				++$index;
				debug "\nAssigning $dataset to storage option $index\n";
				($tmp_storage_assignments_w{$dataset}, $tmp_storage_assignments_rw{$dataset}) = get_storage($persistence_hash{$dataset}, $index);
				if ($tmp_storage_assignments_w{$dataset} eq "") {
					last;
				}
				if ($tmp_storage_assignments_w{$dataset} eq $storage_assignments_w{$dataset} and $tmp_storage_assignments_rw{$dataset} eq $storage_assignments_rw{$dataset}) {
					next;
				}

				provision_datasets_tasks(\%tmp_storage_assignments_rw, \%tmp_storage_assignments_w, \%parallelism);
				debug "\nTrying out configuration:\n";
				if ($DEBUG) {
				display_assignment(\%tmp_storage_assignments_rw, \%tmp_storage_assignments_w, \%parallelism);
				}
				my $cost = provision_cluster_wrapper($min_cost, 0, \%tmp_storage_assignments_rw, \%tmp_storage_assignments_w, \%parallelism);
				debug "Cost = $cost\n";
				if ($cost < $min_cost) {
					debug "Updating min_cost from $min_cost to $cost\n";
					@min_cost_perturbation = ($dataset, $tmp_storage_assignments_w{$dataset}, $tmp_storage_assignments_rw{$dataset});
					$min_cost = $cost;
				}
			}
		}

		if ($#min_cost_perturbation < 0) {
			last;
		}
		$sys_cost = $min_cost;
		my $dataset = $min_cost_perturbation[0];
		$storage_assignments_w{$dataset} = $min_cost_perturbation[1];
		$storage_assignments_rw{$dataset} = $min_cost_perturbation[2];
	}

	provision_datasets_tasks(\%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
	$sys_cost = provision_cluster_wrapper($INFINITY, 0, \%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
	return $sys_cost;
}

sub display_assignment {
	my ($storage_assignments_rw, $storage_assignments_w, $parallelism) = @_;

	foreach my $dataset (keys %cap_hash) {
		if (not(exists($remote_dataset_hash{$dataset}))) {
			next;
		}
		printf "Storage for $dataset = %d units of %s with total IO = %d", ceil(${$storage_reqs_rw{$dataset}}[0]), ${$storage_assignments_rw->{$dataset}}{"name"}, ${$storage_reqs_rw{$dataset}}[1] * ${$storage_reqs_rw{$dataset}}[0];
		if ($storage_assignments_w->{$dataset} ne $storage_assignments_rw->{$dataset}) {
			printf "(in addition, %d units of %s with total IO = %d)", ceil(${$storage_reqs_w{$dataset}}[0]), ${$storage_assignments_w->{$dataset}}{"name"}, ${$storage_reqs_w{$dataset}}[1] * ${$storage_reqs_w{$dataset}}[0];
		}
		print "\n";
	}
	foreach my $process (keys %process_sets) {
		printf "Requirements for $process = %d CPUs and parallelism of %d", ceil($compute_reqs{$process}), $parallelism->{$process};
		print "\n";
		foreach my $dataset (@{$process_locality_hash{$process}}) {
			printf "Storage for $dataset per instance of $process = %d units of %s with total IO = %d", ceil(${$storage_reqs_rw{$dataset}}[0]), ${$storage_assignments_rw->{$dataset}}{"name"}, ${$storage_reqs_rw{$dataset}}[1] * ${$storage_reqs_rw{$dataset}}[0];
			if ($storage_assignments_w->{$dataset} ne $storage_assignments_rw->{$dataset}) {
				printf "(in addition, %d units of %s with total IO = %d)", ceil(${$storage_reqs_w{$dataset}}[0]), ${$storage_assignments_w->{$dataset}}{"name"}, ${$storage_reqs_w{$dataset}}[1] * ${$storage_reqs_w{$dataset}}[0];
			}
			print "\n";
		}
	}
}

my %min_storage_assignments_rw = ();
my %min_storage_assignments_w = ();
my %min_parallelism = ();

my %costs_hash = ();
my $min_cost = $INFINITY;
for (my $i = 0; $i < $NUM_SAMPLES; ++$i) {
	my $sys_cost = run_model();
	if ($sys_cost == $INFINITY) {
		--$i;
		next;
	}
	$costs_hash{$sys_cost} = 1;
	if ($sys_cost < $min_cost) {
		$min_cost = $sys_cost;

		%min_storage_assignments_rw = ();
		%min_storage_assignments_w = ();
		%min_parallelism = ();
		foreach my $key (keys %storage_assignments_rw) {
			$min_storage_assignments_rw{$key} = $storage_assignments_rw{$key};
		}
		foreach my $key (keys %storage_assignments_w) {
			$min_storage_assignments_w{$key} = $storage_assignments_w{$key};
		}
		foreach my $key (keys %parallelism) {
			$min_parallelism{$key} = $parallelism{$key};
		}
	}
	if ($silent_mode != 1) {
		print "Cost sample $i: $sys_cost\n";
		display_assignment(\%storage_assignments_rw, \%storage_assignments_w, \%parallelism);
	}
}
print "Minimum cost: $min_cost\n";
display_assignment(\%min_storage_assignments_rw, \%min_storage_assignments_w, \%min_parallelism);
my @costs = keys %costs_hash;
printf "No. of distinct cost samples = %d\n", ($#costs + 1);
